/*******************************************************************************

    METAL DETECTOR PINPOINTER_LFM

    COM INDICAO PARA METAL FERROSO E NO FERROSO POR MEIO DE 2 LEDS

    C/ 6 LEDS ADICIONAIS FORMANDO UM BARGRAPH DE GRU DE PROXIMIDADE

    ATMEGA8_16PU

    AUTOR: CLAUDIO LARIOS

    DATA:25/02/2024

    COMPILADOR MIKROC PRO AVR V.6.1.0

    USO DIDTICO
    ESTE ARQUIVO  PARTE INTEGRANTE DO BLOG PICSOURCE

*******************************************************************************/
  //prottipos de subrotinas
  void delay10ms(void);
  void delay100ms(void);
  void initial_config(void);
  void var_delay_us(char v);
  void gera_beep(char d,char e);
  void beep_erros(char n);
  void nonferrous();
  void ferrous();
  void dinamic_adjust_value();
  void efeito_sonoro_taranra();
  void detecta_metal();
  void testa_osc_sensor();
  void oscillator_stability();

    //BITS
sbit  lednf   at PORTB2_bit;
sbit  ledf    at PORTB3_bit;
sbit  buzzer  at PORTB1_bit;
sbit  led0    at PORTD0_bit;
sbit  led1    at PORTD1_bit;
sbit  led2    at PORTD2_bit;
sbit  led3    at PORTD3_bit;
sbit  led4    at PORTD5_bit;
sbit  led5    at PORTD6_bit;


sbit  lednf_Direction    at DDB2_bit;
sbit  ledf_Direction     at DDB3_bit;
sbit  buzzer_Direction   at DDB1_bit;
sbit  led0_Direction     at DDD0_bit;
sbit  led1_Direction     at DDD1_bit;
sbit  led2_Direction     at DDD2_bit;
sbit  led3_Direction     at DDD3_bit;
sbit  led4_Direction     at DDD5_bit;
sbit  led5_Direction     at DDD6_bit;

//entrada de sinal da bobina osciladora detetora no pino 5 (t0cki)
//==============================================================================

//variveis globais
bit flag_ok;//flag indicador de nova leitura do timer 1 durante interrupo do tmr0
bit flag_g1;//flag geral 1
bit flag_g2;//flag geral 2
bit flag_fnf;//flag indicador de metal ferroso ou no ferroso
bit flag_ovft0;//flag auxiliar indicador de ocorrncia de estouro no timer 0

char conta_ovf1=0; //contador de overflow do timer 1
char salva_overf1=0;//salva a contagem do overflow do timer 1
char salva_tmr0=0;//salva contagem do timer 0
char salva_tmr1h;//salva contagem do timer 1 H
char salva_tmr1l;//salva contagem do timer 1 L
char cnt1=0;//contador do filtro 1
char cnt2=0;//contador do filtro 2
char cnt_prox=0;//guarda valor de proximidade
long new_tmr1;//guarda a ltima leitura do timer 1
long old_tmr1a=0;//guarda a leitura antiga principal do timer 1
long old_tmr1b=0;//guarda copia da leitura antiga para uso no filtro
long old_tmr1c=0;//guarda copia da leitura antiga para uso no filtro
long dif=0;//guarda a diferena entre as leituras atuais
long old_dif=0;//guarda a diferena de leituras anteriores
int cnt_it=0; //contador de iteraes mximas
 /************************************************************************************
  Rotina de interrupo do timer 1
  Objetivo da rotina: Incrementar contador adicional quando ocorrer overflow
  do timer 1.
/***********************************************************************************/
 void TIMER1_ISR() iv IVT_ADDR_TIMER1_OVF ics ICS_AUTO {

   conta_ovf1++;  //conta quantos estouros do timer 1 ocorreu.
}
/************************************************************************************
  Rotina de interrupo do timer 0
  Objetivo da rotina: Obter o valor da frequncia da bobina sensora ligada
  na entrada (pino 5 -configurado como entrada do timer 0). As alteraes entre as
  medies sero usadas para determinar a presena de metal ferroso  ou no ferroso.
  O timer 0 gera na realidade apenas o intervalo de tempo. O timer 1 conta este
  intervalo entre a interrupes, sendo este proporcional a frequncia medida.
/***********************************************************************************/
void TIMER0_ISR() iv IVT_ADDR_TIMER0_OVF ics ICS_AUTO {
int a,b;

    b=TCNT1L;  //parte L tem que ler primeiro
    a=TCNT1H;
    salva_tmr1h=a;//salva medio para uso nas rotinas
    salva_tmr1l=b;
    TCNT1H=0;//parte h=0  tem que ser escrita primeiro
    TCNT1L=0;//zera timer1 h e timer 1l para nova medio
    salva_overf1=conta_ovf1;
    conta_ovf1=0;
    TOV1_bit=1;//reseta flag over flow do timer1  com '1'
    flag_ok=1;
    TCNT0=salva_tmr0;//recarga do timer 0
}
 /************************************************************************************
  Rotina de delay fixa em 10 milissegundos
/***********************************************************************************/
void delay10ms(){

  asm WDR ; //reseta WDT
  delay_ms(10);
  asm WDR ;
}
/************************************************************************************
  Rotina de delay fixa em 100 milissegundos
/***********************************************************************************/
void delay100ms(){
 char a;

   for(a=0;a<10;a++)delay10ms();
}
/************************************************************************************
  Rotina de setup inicial de registradores e ports
/***********************************************************************************/

void initial_config(){

   TCCR0=0b00000111; //timer0 conta com pulso externos do pino 8 (PD4)
   ACSR=0X80; //desliga comparador (economia de energia)
   DDRB=0B00001111;//todos os pinos como sada
   DDRD=0B11101111;//entradas e sadas
   PORTB=0;
   PORTD=0;
   flag_ok=0;//flag indicador de nova leitura do timer 1 durante interrupo do tmr0
   flag_g1=0;//flag geral 1
   flag_g2=0;//flag geral 2
   flag_fnf=0;//flag indicador de metal ferroso ou no ferroso
   flag_ovft0=0;//flag auxiliar indicador de ocorrncia de estouro no timer 0
   delay100ms();//delay total de 300 ms
   delay100ms();
   delay100ms();
   SREG_I_bit = 0;  // Desliga geral das interrupes
   TCNT0=0;//zera timer 0
   asm WDR ;//RESETA WDT
   WDTCR=0B00001010; //WDT com 60 ms
   TCCR1A=0;
   TCCR1B=1; //timer 1 sem prescaller sinal clock interno
   TIMSK=0B00000001;//Liga interrupo de overflow do timer0
}
/************************************************************************************
  Rotina de delay varivel em microssegundos.
  Objetivo da rotina: Obter delay de tempos reduzidos com restart do wdt.
/***********************************************************************************/

void var_delay_us(char v){

  if(v>=3)v=v-3; else return;
  while(v){
   v--;
   asm NOP;
   asm NOP;
   asm NOP;
   asm NOP;
   asm NOP;
   asm NOP;
   asm NOP;
   asm NOP;
   asm NOP;
   asm WDR ;
  }
}
/************************************************************************************
  Rotina de gerao de som
  Objetivo da rotina: Obter sons (apitos) com durao determinada por 'd' e frequncia
  sonoro determinada por 'e'.
/***********************************************************************************/
void gera_beep(char d,char e){
  char c;

     PORTD=0;
     ledf=0;
     lednf=0;
     for(c=0;c<d;c++){
       buzzer=1;
       var_delay_us(e);
       buzzer=0;
       var_delay_us(e);
     }//for(c=0;c<=d;c++){
 }//void gera_beep(d,e){

/************************************************************************************
  Rotina de gerao de efeito sonoro.
  Objetivo da rotina: Gerar indicao de erros no  hardware por meio de beeps.
  Sendo o efeito sonoro comeando com beep longo, um delay 1 seg , mais outros
  beeps cuja quantidade indica qual  o erro, seguido de delay de 3 segs.

  1 beep = oscilador com frequncia abaixo de 18 khz
  2 beeps = frequncia acima de 22 khz
  4 beeps = oscilador morto
/***********************************************************************************/

void beep_erros(char n){
char d;

  for(d=0;d<8;d++){  // beep longo
   gera_beep(255,255);
  }
  for(d=0;d<10;d++){//1 segundo de delay
    delay100ms();
  }
  for(d=0;d<n;d++){
    gera_beep(230,150);
    delay100ms();
    delay100ms();
    delay100ms();
    delay100ms();
  }
  for(d=0;d<30;d++){//delay de 3 segundos
   delay100ms();
  }
    for(;;);  //Espera reset do WDT
}//beep_erros()

/************************************************************************************
  Rotina de indicao de presena de metais no-ferrosos
  Objetivo da rotina: gerar som distinto para metal no ferroso e acender led nf (verde)
/***********************************************************************************/

void nonferrous(){
char a;

  lednf=1;
  ledf=0;
  for(a=0;a<=150;a++){
    buzzer=1;
    var_delay_us(130);//110
    buzzer=0;
    var_delay_us(130);//110
  }
  delay10ms();
  lednf=0;
  ledf=0;
}
/************************************************************************************
  Rotina de indicao de presena de metais ferrosos
  Objetivo da rotina: gerar som distinto para metal ferroso e acender led f (azul)
/***********************************************************************************/
void ferrous(){
char a;

  lednf=0;
  ledf=1;
  for(a=0;a<=80;a++){
    buzzer=1;
    var_delay_us(250);
    buzzer=0;
    var_delay_us(250);
  }
  delay10ms();
  lednf=0;
  ledf=0;
}
/************************************************************************************
  Rotina de ajuste dinmico do valor de comparao de leituras do timer 1
  Objetivo da rotina: Ajustar o valor de comparao compensando pequenas variaes
  da frequncia do oscilador.  Em outras palavras o valor de comparao 'tende'
  lentamente a ser igual a leitura atual do timer 1.
/***********************************************************************************/

void dinamic_adjust_value(){
 #define FATOR1 15
 #define FATOR2 15

  asm WDR ;
  if(old_dif > dif) old_tmr1b=old_tmr1a;//se diferenca anterior  maior que diferena atual
  old_dif=dif;//diferena anterior fica com diferena atual

  if(old_tmr1c>new_tmr1)old_tmr1c--;
  if(old_tmr1c<new_tmr1)old_tmr1c++;//se valor anterior timer1 c for menor que nova leitura timer1

  cnt1++;
  if(cnt1>=FATOR1){//rotina realizada a cada FATOR 1 iteraes
    cnt1=0;
    cnt2++;
    if(old_tmr1c>old_tmr1b)old_tmr1b++;//c>b=b++
    if(old_tmr1c<old_tmr1b)old_tmr1b--;//c<b=b--
  }
  if(cnt2>=FATOR2){//rotina realizada a cada FATOR2 iteraes
    cnt2=0;
    if(old_tmr1a>old_tmr1b)old_tmr1a--;//a>b=a--
    if(old_tmr1a<old_tmr1b)old_tmr1a++;//a<b=a++
  }
}//dinamic_adjust_value(){

/************************************************************************************
  Rotina de gerao de som
  Objetivo da rotina:
  Gerar beep de som caracteristico (ta-r-r)
/***********************************************************************************/

void efeito_sonoro_taranra(){
  char a;

  PORTD=0; //leds apagados
  lednf=0;
  ledf=0;
  for(a=0;a<3;a++){//gera beep 1
  gera_beep(100,240);//n. rep/periodo da nota
  }
  for(a=0;a<3;a++){
  delay10ms();//delay de 30 ms
  }
  for(a=0;a<3;a++){//gera beep 2
  gera_beep(200,120);//n. rep/periodo da nota
  }
  for(a=0;a<3;a++){
  delay10ms();//delay de 30 ms
  }
  for(a=0;a<3;a++){///gera beep 3
  gera_beep(120,200);//n. rep/periodo da nota
  }
}
/************************************************************************************
  Rotina de deteco de metal
  Objetivo da rotina:
  1) compara as diferenas de leitura anterior e atual
  2) decidir se leitura  suficiente para acionar led/beep
  3) decidir qual led acender e frequncia do beep
  4) obter retardo de tempo proporcional a amplitude da diferena nas leituras:
     (usando mtodo de delocamento e contador de grau de proximidade de 3 nveis)
/***********************************************************************************/

void detecta_metal(){
  int aa,bb;
  char c,proximity;

  #define N_ITERACOES  12000
  #define N_MIN  4    //entre 3 e 10

  asm WDR ;
  flag_ok=0;//reseta flag
  while(!flag_ok);//aguarda ficar '1' indicando nova leitura da frequencia osc.
  new_tmr1=(long)salva_tmr1l | (long)salva_tmr1h<<8 | (long) salva_overf1<<16;
  asm WDR ;
  if(new_tmr1>old_tmr1a){
    dif=new_tmr1 - old_tmr1a;//frequencia nova maior que anterior
    flag_fnf=0;//indica metal no ferroso
  }
 else if(new_tmr1==old_tmr1a){//sem alterao da frequncia (sem presena de metal no sensor)
   dif=0;
   flag_fnf=0;
 }
 else if(new_tmr1<old_tmr1a){//frequencia nova menor que anterior
    dif=old_tmr1a - new_tmr1;
    flag_fnf=1;//indica metal ferroso
 }

 if(dif<=N_MIN){    //se diferena de leituras forem menor que N_MIN
   cnt_prox=1;
   cnt_it++;  //incrementa contador de iteraes
   if(cnt_it>N_ITERACOES){  //testa se chegou a 12000 iteraes
    cnt_it=0;        //sim,zera contador de iteraes
    efeito_sonoro_taranra();    //chama rotina de som caracteristica
   }
   dinamic_adjust_value();  //rotina de autoajuste do valor de comparao da freq do osc.ext
 }//if(dif<N_MIN){

 else{//else - if(dif<N_MIN){
 //se diferena de leituras forem maior ou igual a 3
   cnt_it=0;
   //rotina de 3 nveis de indicao de aproximao
   // Objetivo: obter retardo proporcional ao valor da diferena de leitura
   //alterando a velocidade do beep e piscada dos led f/nf em 3 nveis
    aa=dif;
    bb=0x2000;
    for(c=0;c<=13;c++){//14 - em 'c' se obter um valor de 0 a 13
      asm WDR ;
      if(aa & bb)break;
      bb=bb>>1;
    }
    proximity=c;
    proximity=proximity>>2; //2  'c' ser dividido por 4 para obter apenas 3 nveis de visualizao

    if(proximity>=cnt_prox){//verifica se  a vez de acionar a sada conforme proximidade
      cnt_prox++;
    }// if(proximity>cnt_prox){
   else{
     cnt_prox=1;
     portb=0;
     portd=0;
     c=c>>1;
     if(c>5)c=5;

       //obtem o efeito de bargraph com os leds 0 a 5
     switch (c){

      case 0: led5=1;
      case 1: led4=1;
      case 2: led3=1;
      case 3: led2=1;
      case 4: led1=1;
      case 5: led0=1;
     }

     if(flag_fnf==1){
       nonferrous();//apita e acende led nonferrous (led verde)
     }
     else{
       ferrous();//apita e acende led ferrous (led azul)
     }
     portb=0; //desliga leds do bargraph
     portd=0;
   }//else -  if(proximity>cnt_prox){

 }//else - if(dif<N_MIN){

}//void detecta_metal(){

/************************************************************************************
  Rotina de teste do oscilador da bobina sensora de metais
  Objetivo: Ajusta valor de recarga do timer0
  Caso o oscilador no esteja funcionando ou com frequncia acima ou abaixo
  ser feito a indicao por meio de piscadas e apitos.
/***********************************************************************************/

void testa_osc_sensor(){
 char new_tmr0;

    SREG_I_bit = 0;  // Desliga geral das interrupes
    nonferrous();
    delay100ms();
    delay100ms();
    TCNT0=0;
    TOV0_bit=1;//reseta flag over flow do timer0  com '1'
    delay10ms();
    new_tmr0=TCNT0;
    flag_ovft0=0;
    if(TOV0_bit)flag_ovft0=1;//houve over flow do timer0
    TOV0_bit=1; //apaga flag timer 0 com '1'
    asm WDR ;
    salva_tmr0=new_tmr0;//pega valor do timer 0 salvo

     if(250<new_tmr0){beep_erros(2);}//freq>22khz, vai daqui para inicio do programa
     else{//f
       if(flag_ovft0==1){beep_erros(2);}//freq>22khz, vai daqui para inicio do programa
       else{
         if(2<new_tmr0){
          if(149>=new_tmr0){beep_erros(1);}//freq< 18khz, vai daqui para inicio do programa
         }
         flag_g1=1;
       }//x
     }//else{//f

   if((new_tmr0==0)&& (flag_ovft0==0)){
     beep_erros(4); //sem sinal pino 5, apito longo + pausa + 4x e goto inicio
   }
   salva_tmr0 = 0xff - new_tmr0;
}//void testa_osc_sensor(){

/************************************************************************************
  Rotina de estabilizao do oscilador ligado ao pino 5 (do osc. sensor de metal)
  Objetivo da rotina: Aguardar a estabilizao das leituras do timer 1
  (sem presena de metais na bobina do sensor do oscilador)
/***********************************************************************************/

void oscillator_stability(){
unsigned char a,f,g,i;

  flag_g2=0;//reseta flag
  f=0;
  g=0;
  TCNT1L=0; //limpa timer1 l
  TCNT1H=0; //limpa timer1 h
  TCNT1L=0; //limpa timer1 l
  TOV1_bit=1;//reseta flag over flow do timer1  com '1'
  TIMSK=0B00000101;//Liga interrupo de overflow do timer0 e timer1
  TOV0_bit=1;//reseta flag over flow do timer0  com '1'
  SREG_I_bit = 1;  // liga geral das interrupes
  flag_ok=0;
  while(!flag_ok);//aguarda ficar '1'
  old_tmr1a=(long)salva_tmr1l | (long)salva_tmr1h<<8 | (long) salva_overf1<<16;

  do{
     asm WDR ;
     flag_ok=0;//reseta flag
     while(!flag_ok);//aguarda ficar '1'
     new_tmr1=(long)salva_tmr1l | (long)salva_tmr1h<<8 | (long) salva_overf1<<16;
     if(new_tmr1 > old_tmr1a) old_tmr1a++;
     if(old_tmr1a > new_tmr1) old_tmr1a--;
     g++;
     if(g==0)beep_erros(3);//se houver mais de 255 iteraes acusa erro
     if(old_tmr1a == new_tmr1){
       f++;
       if(f>3)flag_g2=1;
     }
     else{
       f=0;
       old_tmr1a=new_tmr1;
     }
   }while(flag_g2==0);  // do{

//salva cpias nos registradores
   old_tmr1b=old_tmr1a;
   old_tmr1c=old_tmr1a;
   PORTB=0;
   lednf=0; //leds apagados
   ledf=0;
   i=255;
   do{  //produz breve som (apito) do grave para agudo (tuuuuiiiiii)
     buzzer=1;
     var_delay_us(i);
     buzzer=0;
     var_delay_us(i);
     var_delay_us(i);
     i--;
   }while(i>100);

  for(a=0;a<5;a++)delay100ms();//delay 500 ms
 }

/*******************************************************************************
          ROTINA MAIN
*******************************************************************************/

void main() {

  initial_config(); //configuraes iniciais de ports e registradores
  testa_osc_sensor();//verifica funcionamento do oscilador do sensor de metal
  oscillator_stability(); //aguarda estabilizar o oscilador do sensor de metal

/*******************************************************************************
           LOOP PRINCIPAL DE REPETIO
*******************************************************************************/
   for(;;){
    detecta_metal();//rotina de trabalho para detectar metais ferrosos e no ferrosos
   }//for(;;)
}//void main()